#!/usr/bin/env python

"""
Generate low-level code.

Copyright (C) 2007, 2008, 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.opt import Optimiser
from micropython.common import *
from micropython.data import *
from micropython.rsvp import *
import compiler.ast

class Assembler:

    "Support for generating low-level code."

    def __init__(self, optimisations):

        "Initialise the assembler with an optimiser and status attributes."

        # Optimisation.

        self.optimiser = Optimiser(self, optimisations)

        # The current unit being translated.

        self.unit = None

        # The temporary storage used by the current assignment expression.

        self.expr_temp = []

        # Wiring within the code.

        self.labels = {}
        self.label_number = 0
        self.loop_blocks = []
        self.exception_blocks = []

    def reset(self):

        "Reset the state of the assembler."

        # The code itself. This is limited to the code for a particular block
        # being processed.

        self.blocks = []

        # Information about temporary values.

        self.temp_positions = set()
        self.max_temp_position = -1

        # Information about instructions which construct frames.

        self.frame_makers = []

        # Optimiser state must be reset for each unit.

        self.optimiser.reset()

    def new_block(self):

        "Return a new code block."

        return Block()

    def get_block(self):

        "Return the current block."

        return self.blocks[-1]

    def set_block(self, block, preceding=None):

        """
        Add the given 'block' to the unit's list of blocks, noting any active
        value information from 'preceding' blocks on the new block.
        """

        self.optimiser.reset(block, preceding)
        self.blocks.append(block)

    def get_loop_blocks(self):
        return self.loop_blocks[-1]

    def add_loop_blocks(self, next_block, exit_block):
        self.loop_blocks.append((next_block, exit_block))

    def drop_loop_blocks(self):
        self.loop_blocks.pop()

    def add_exception_unit(self):
        self.exception_blocks.append([])

    def get_exception_blocks(self):
        return self.exception_blocks[-1][-1]

    def add_exception_blocks(self, handler_block, exit_block):
        self.exception_blocks[-1].append((handler_block, exit_block))

    def drop_exception_blocks(self):
        self.exception_blocks[-1].pop()

    def drop_exception_unit(self):
        self.exception_blocks.pop()

    # Assignment expression values.

    def record_value(self, immediate=1):

        """
        Record the current active value for an assignment. If the optional
        'immediate' parameter is set to a false value, new temporary storage to
        hold the recorded value will be allocated; otherwise, the
        value-providing instruction may be replicated in order to provide the
        active value later on.
        """

        if immediate:
            temp = self.optimiser.optimise_temp_storage()
        else:
            temp = self.get_temp()
        self.expr_temp.append(temp)

    def discard_value(self):

        "Discard any temporary storage in use for the current assignment value."

        self.discard_temp(self.expr_temp.pop())

    def assign_value(self, expr=None):

        """
        Set the source of an assignment using 'expr' or the current assignment
        value. This sets the source register of the current instruction.
        """

        if expr is None:
            expr = self.expr_temp[-1]

            # Optimise away constant storage if appropriate.

            if self.optimiser.optimise_constant_storage(expr):
                self.remove_op()
                return

        # Otherwise, insert the assignment source.

        if expr is not None:
            expr_copy = expr.copy()
            assign_op = self.last_op()

            # Either insert the instruction yielding the value and adjust the
            # assignment source.

            expr_copy.target = "source"

            if self.insert_op(-1, expr_copy):
                assign_op.source = "source"
                self.update_temp(expr, expr_copy)

            # (Now, the instruction need not be inserted.)

            # Or transfer the working value to the source register.

            elif assign_op.working == "working":
                self.insert_op(-1, Transfer(source="working_context", target="source_context"))
                self.insert_op(-1, Transfer(source="working", target="source"))
                assign_op.source = "source"

            # Or let the assignment use the working register.

            else:
                assign_op.source = "working"

    def set_target(self, target):

        "Reset the target of the active instruction to 'target'."

        self.optimiser.set_target(target)

    def is_immediate_user(self, node):

        """
        Return whether 'node' is an immediate user of an assignment expression.
        """

        return isinstance(node, (compiler.ast.AssName, compiler.ast.AssAttr))

    def has_immediate_usage(self, nodes):

        """
        Return whether 'nodes' are all immediate users of an assignment expression.
        """

        for n in nodes:
            if not self.is_immediate_user(n):
                return 0
        return 1

    # Temporary storage administration.

    def get_temp(self):

        """
        Return a temporary storage access instruction for the current value.
        Initially, this instruction is not associated with an allocated unit of
        temporary storage, and if used as a new instruction will not be added to
        the code, but if the current value changes, the 'set_temp' method will
        be called by the optimiser and a unit of storage will be allocated.
        """

        temp = LoadTemp(None)
        self.optimiser.request_active_value(temp, self.blocks[-1], len(self.blocks[-1].code))
        return temp

    def set_temp(self, temp, block, pos):

        """
        Emit a storage instruction for the given 'temp' loading instruction,
        reserving a new temporary storage location.
        """

        if temp.attr is None:
            temp.attr = self.reserve_temp()
            block.insert(pos, StoreTemp(temp.attr))

    def update_temp(self, temp, updated):

        "Update 'temp' using the given 'updated' instruction."

        if isinstance(temp, LoadTemp):
            temp.attr = updated.attr

    def reserve_temp(self, temp_position=None):

        """
        Reserve a new temporary storage position, or if the optional
        'temp_position' is specified, ensure that this particular position is
        reserved.
        """

        if temp_position is not None:
            pass
        elif not self.temp_positions:
            temp_position = 0
        else:
            temp_position = max(self.temp_positions) + 1

        self.temp_positions.add(temp_position)
        self.max_temp_position = max(self.max_temp_position, temp_position)
        return self.unit.all_local_usage + temp_position # position in frame

    def ensure_temp(self, instruction=None):

        """
        Ensure that the 'instruction' is using a reserved temporary storage
        position.
        """

        if isinstance(instruction, LoadTemp):
            temp_position = instruction.attr - self.unit.all_local_usage
            self.reserve_temp(temp_position)

    def discard_temp(self, instruction=None):

        "Discard any temporary storage position used by 'instruction'."

        if isinstance(instruction, LoadTemp) and instruction.attr is not None:
            temp_position = instruction.attr - self.unit.all_local_usage
            self.free_temp(temp_position)
            self.optimiser.ignore_active_value()

    def free_temp(self, temp_position):

        "Free the temporary storage position specified by 'temp_position'."

        if temp_position in self.temp_positions:
            self.temp_positions.remove(temp_position)

    def set_frame_usage(self, node, extend):

        """
        Ensure that the frame usage for the unit associated with 'node' is set
        on the 'extend' instruction.
        """

        # Remove any ExtendFrame instructions which do nothing.

        if self.last_op() is extend:
            self.remove_op()
            return

        ntemp = self.max_temp_position + 1
        extend.attr = ntemp + node.unit.local_usage # NOTE: See get_code for similar code.

    # Code writing methods.

    def new_op(self, op):

        """
        Add 'op' to the generated code, returning a true value if an instruction
        was added.
        """

        if not self.check_op(op):
            return 0

        # Add the operation to the current block.

        self.optimiser.set_new(op)
        self.blocks[-1].append(op)
        return 1

    def insert_op(self, i, op):

        "Insert at index 'i' in the current block the instruction 'op'."

        if not self.check_op(op):
            return 0

        self.blocks[-1].insert(i, op)
        return 1

    def check_op(self, op):

        "Return whether the given 'op' is to be added to the code."

        # Optimise away temporary storage instructions where the active value is
        # still available and was not recorded.

        if isinstance(op, LoadTemp) and op.attr is None:
            return 0

        # Optimise load operations employed by this instruction.

        if self.optimiser.optimise_away_no_operations(op) or self.optimiser.optimise_unused_handlers(op):
            return 0

        return 1

    def remove_op(self):

        "Remove the last instruction."

        op = self.blocks[-1].code.pop()
        self.optimiser.clear_active()

    def replace_op(self, op):

        "Replace the last added instruction with 'op'."

        self.remove_op()
        self.new_op(op)

    def replace_active_value(self, op):

        """
        Replace the value-providing active instruction with 'op' if appropriate.
        """

        self.optimiser.remove_active_value()
        self.new_op(op)

    def last_op(self):

        "Return the last added instruction."

        try:
            return self.blocks[-1].code[-1]
        except IndexError:
            return None

# vim: tabstop=4 expandtab shiftwidth=4
